import json
import time
import os

import irida_uploader_cl.config as config
from irida_uploader_cl.model.directory_status import DirectoryStatus

from irida_uploader_cl.progress import exceptions


# Module level Constants
# These define the status files valid fields and variables

# File name
STATUS_FILE_NAME = "irida_uploader_status.info"

# Status field for a sequencing run
STATUS_FIELD = "Upload Status"
DATE_TIME_FIELD = "Date Time"
RUN_ID_FIELD = "Run ID"
IRIDA_INSTANCE_FIELD = "IRIDA Instance"
MESSAGE_FIELD = "Message"


def get_directory_status(directory, required_file_list):
    """
    Gets the directory status based off using 'irida_uploader_status.info' files to track progress

    :param directory: the directory to search for a run
    :param required_file_list: optional param: a list of required files that
        are required for that run to be considered valid. Example: ['SampleSheet.csv']
    :return: directory and status dictionary
    """
    result = DirectoryStatus(directory)

    if not os.access(directory, os.W_OK):
        result.status = DirectoryStatus.INVALID
        result.message = 'Directory cannot be accessed. Please check permissions'
        return result

    file_list = next(os.walk(directory))[2]  # Gets the list of files in the directory

    # Legacy upload catch
    # When the irida-miseq-uploader (old uploader) ran it generated a .miseqUploaderInfo file
    # To prevent uploading runs that used this old system, we assume runs with this file are COMPLETE
    # By default they will not be picked up automatically with --batch because they are set to COMPLETE,
    # but they can still be uploaded using the --force option
    if '.miseqUploaderInfo' in file_list:
        result.status = DirectoryStatus.COMPLETE
        result.message = "Legacy uploader run. Set to complete to avoid uploading duplicate data."
        return result

    for file_name in required_file_list:
        if file_name not in file_list:
            result.status = DirectoryStatus.INVALID
            result.message = 'Directory is missing required file with filename {}'.format(file_name)
            return result

    if STATUS_FILE_NAME not in file_list:  # no irida_uploader_status.info file yet, has not been uploaded
        result.status = DirectoryStatus.NEW
        return result

    # Must check status of upload to determine if upload is completed
    uploader_info_file = os.path.join(directory, STATUS_FILE_NAME)
    with open(uploader_info_file, "rb") as reader:
        data = reader.read().decode()
    info_file = json.loads(data)
    status = info_file[STATUS_FIELD]
    if status in DirectoryStatus.VALID_STATUS_LIST:
        result.status = status
        if MESSAGE_FIELD in info_file:
            result.message = info_file[MESSAGE_FIELD]
        else:
            result.message = None
    else:  # the status found in the file is not in the defined list
        raise exceptions.DirectoryError("Invalid Status in status file", directory)

    return result


def write_directory_status(directory_status, run_id=None):
    """
    Writes a status to the status file:
    Overwrites anything that is in the file

    Writes a timestamp to the time of last written

    :param directory_status: DirectoryStatus object containing status to write to directory
    :param run_id: optional, when used, the run id will be included in the status file,
        along with the irida instance the run is uploaded to.
    :return: None
    """

    if not os.access(directory_status.directory, os.W_OK):  # Cannot access upload directory
        raise exceptions.DirectoryError("Cannot access directory", directory_status.directory)

    uploader_info_file = os.path.join(directory_status.directory, STATUS_FILE_NAME)
    if run_id:
        json_data = {STATUS_FIELD: directory_status.status,
                     MESSAGE_FIELD: directory_status.message,
                     DATE_TIME_FIELD: _get_date_time_field(),
                     RUN_ID_FIELD: run_id,
                     IRIDA_INSTANCE_FIELD: config.read_config_option('base_url')}
    else:
        json_data = {STATUS_FIELD: directory_status.status,
                     MESSAGE_FIELD: directory_status.message,
                     DATE_TIME_FIELD: _get_date_time_field()}

    with open(uploader_info_file, "w") as json_file:
        json.dump(json_data, json_file, indent=4, sort_keys=True)
        json_file.write("\n")


def _get_date_time_field():
    """
    Returns the current date and time as a string
    :return:
    """
    return time.strftime("%Y-%m-%d %H:%M")



from irida_uploader_cl.model.sequencing_run import SequencingRun
from irida_uploader_cl.model.project import Project
from irida_uploader_cl.model.sample import Sample
from irida_uploader_cl.model.sequence_file import SequenceFile
from irida_uploader_cl.model.directory_status import DirectoryStatus
from irida_uploader_cl.model.validation_result import ValidationResult
from irida_uploader_cl.model import exceptions



#!/usr/bin/env python3

import argparse
import getpass
import os
import textwrap

import os
module_dir=os.path.dirname(os.path.abspath(__file__))
for module in module_dir:
    if module == '__init__.py' or module[-3:] != '.py':
        continue
    __import__(module[:-3], locals(), globals())
del module

import irida_uploader_cl.config as config
import irida_uploader_cl.core as core
from irida_uploader_cl.parsers import supported_parsers

# Set up an argument parser. We are using defaults to stay consistent with other software.
# description gets added to the usage statements
argument_parser = argparse.ArgumentParser(
    formatter_class=argparse.RawDescriptionHelpFormatter,
    description=textwrap.dedent('''
    This program parses sequencing runs and uploads them to IRIDA.

    required arguments:
      --d DIRECTORY, --directory DIRECTORY
                            Location of sequencing run to upload.
                            Directory must be writable.
    '''),
    prog="irida-uploader.sh -d DIRECTORY",
    epilog='-c* options can be used without a parameter to prompt for input.')
# Our main argument. It is required or else an error will be thrown when the program is run
# Normally we would use a positional argument, but because of our 1 or None config overrides it makes sense to use
# a optional argument, with the required set to True. We have to suppress the help on the argument, and add it's help
# information to the formatted description text of the parser for the argument to be shown as required when --help
# is used.
argument_parser.add_argument('-d', '--directory',
                             action='store',
                             required=True,
                             help=argparse.SUPPRESS)
# help='Location of sequencing run to upload. Directory must be writable.')
# Add the version argument
argument_parser.add_argument('--version',
                             action='version', version='IRIDA Uploader {}'.format(core.VERSION_NUMBER))
# Optional argument, for using an alternative config file.
argument_parser.add_argument('-c', '--config',
                             action='store',
                             help='Path to an alternative configuration file. '
                                  'This overrides the default config file in the config directory')
# Optional argument, Force uploading a run even if a non new status file exists
argument_parser.add_argument('-f', '--force',
                             action='store_true',  # This line makes it not parse a variable
                             help='Uploader will ignore the status file, '
                                  'and try to upload even when a run is in non new status.')
# Optional argument, Upload all sequencing runs in a directory of runs
argument_parser.add_argument('-b', '--batch',
                             action='store_true',  # This line makes it not parse a variable
                             help='Uploader will expect a directory containing a sequencing run directories, '
                                  'and upload in batch. '
                                  'The list of runs is generated at start time '
                                  '(Runs added to directory mid upload will not be uploaded).')

# Optional arguments for overriding config file settings
# Explanation:
#   nargs='?', const=True, default=False,
#       Allows zero or one parameters
#       If the argument is not given:                      the value will be False           (indicates load from file)
#       If the argument is given, and no parameter given:  the value will be True            (prompt user for input)
#       If the argument is given, and parameter is given:  the value will be the parameter   (used to override config)
argument_parser.add_argument('-ci', '--config_client_id', action='store', nargs='?', const=True, default=False,
                             help='Override the "client_id" field in config file. '
                                  'This is for the uploader client created in IRIDA, used to handle the data upload')
argument_parser.add_argument('-cs', '--config_client_secret', action='store', nargs='?', const=True, default=False,
                             help='Override "client_secret" in config file. '
                                  'This is for the uploader client created in IRIDA, used to handle the data upload')
argument_parser.add_argument('-cu', '--config_username', action='store', nargs='?', const=True, default=False,
                             help='Override "username" in config file. This is your IRIDA account username.')
argument_parser.add_argument('-cp', '--config_password', action='store', nargs='?', const=True, default=False,
                             help='Override "password" in config file. This corresponds to your IRIDA account.')
argument_parser.add_argument('-cb', '--config_base_url', action='store', nargs='?', const=True, default=False,
                             help='Override "base_url" in config file. The api url for your irida instance '
                                  '(example: https://my.irida.server/api/)')
argument_parser.add_argument('-cr', '--config_parser', action='store', nargs='?', const=True, default=False,
                             help='Override "parser" in config file. '
                                  'Choose the type of sequencer data that is being uploaded. '
                                  'Supported parsers: ' + str(supported_parsers))


def _set_config_override(args):
    """
    Check the config override arguments and override
    :param args: list of args from parseargs
    :return:
    """
    client_id = None
    client_secret = None
    username = None
    password = None
    base_url = None
    parser = None

    if args.config_client_id is True:
        print("Enter Client ID:")
        client_id = input()
    elif args.config_client_id is not False:
        client_id = args.config_client_id

    if args.config_client_secret is True:
        print("Enter Client Secret:")
        # getpass blanks out the secret entry
        client_secret = getpass.getpass()
    elif args.config_client_secret is not False:
        client_secret = args.config_client_secret

    if args.config_username is True:
        print("Enter Username:")
        username = input()
    elif args.config_username is not False:
        username = args.config_username

    if args.config_password is True:
        print("Enter Password:")
        # getpass blanks out password entry
        password = getpass.getpass()
    elif args.config_password is not False:
        password = args.config_password

    if args.config_base_url is True:
        print("Enter Base IRIDA URL (format: http://my.irida.server/api/):")
        base_url = input()
    elif args.config_base_url is not False:
        base_url = args.config_base_url

    if args.config_parser is True:
        print("Enter Parser to use:")
        parser = input()
    elif args.config_parser is not False:
        parser = args.config_parser

    config.set_config_options(client_id=client_id,
                              client_secret=client_secret,
                              username=username,
                              password=password,
                              base_url=base_url,
                              parser=parser)


def _config_uploader(args):
    """
    Sets up config settings for command line uploading
    :param args:
    :return:
    """
    # If a config file is passed in, set it before starting upload
    if args.config:
        config.set_config_file(args.config)
        # config.user_config_file = args.config
    # Init config
    config.setup()
    # Override with any passed in options
    _set_config_override(args)


def main():
    # Parse the arguments passed from the command line and start the upload
    args = argument_parser.parse_args()
    # Setup config options
    _config_uploader(args)

    # Verify directory is writable before upload
    if not os.access(args.directory, os.W_OK):  # Cannot access upload directory
        print("ERROR! Specified directory is not writable: {}".format(args.directory))
        return 1


    # Start Upload
    if args.batch:
        return upload_batch(args.directory, args.force)
    else:
        return upload(args.directory, args.force)


def upload(run_directory, force_upload):
    """
    start upload on a single run directory
    :param run_directory:
    :param force_upload:
    :return: exit code 0 or 1
    """
    return core.cli_entry.upload_run_single_entry(run_directory, force_upload).exit_code


def upload_batch(batch_directory, force_upload):
    """
    Start uploading runs in the batch directory
    :param batch_directory:
    :param force_upload:
    :return: exit code 0 or 1
    """
    return core.cli_entry.batch_upload_single_entry(batch_directory, force_upload).exit_code


# This is called when the program is run for the first time
if __name__ == "__main__":
    main()



class IridaKeyError(KeyError):
    """
    This error will only happen when there is an issue with the api itself
    It should not be able to happen based on user input

    This could happen if the HATEOAS spec changes,
    or the api has a bug in the implementation of it
    """
    pass



from irida_uploader_cl.api.exceptions.irida_connection_error import IridaConnectionError
from irida_uploader_cl.api.exceptions.irida_key_error import IridaKeyError
from irida_uploader_cl.api.exceptions.irida_resource_error import IridaResourceError
from irida_uploader_cl.api.exceptions.irida_upload_canceled_exception import IridaUploadCanceledException
from irida_uploader_cl.api.exceptions.file_error import FileError



class DirectoryStatus:
    """
    When looking for runs in a directory (or list of directories) this object is used to
    contain the relevant data.
    """

    # States that are valid for the status field

    # New runs, ready to upload
    NEW = 'new'
    # Used when a run directory does not have the base requirements to act as a run
    # Never written to a status file
    INVALID = 'invalid'
    # Parsing/Upload has started/partially completed for this run
    PARTIAL = 'partial'
    # Parsing/Upload has stopped because of some error.
    ERROR = 'error'
    # Parsing/Upload has completed successfully
    COMPLETE = 'complete'

    VALID_STATUS_LIST = [
        NEW,
        INVALID,
        PARTIAL,
        ERROR,
        COMPLETE
    ]

    def __init__(self, directory):
        """
        :param directory: Directory of a potential run
        status: status of the directory: 'new', 'partial', 'complete', 'invalid'
        message: Used when run is invalid,
        """
        self._directory = directory
        self._status = None
        self._message = None

    @property
    def directory(self):
        return self._directory

    @property
    def status(self):
        """
        status property, setter is overridden by @status.setter
        :return: status
        """
        return self._status

    @status.setter
    def status(self, status):
        """
        Status setter
        Only accepts status's from the predefined VALID_STATUS_LIST
        :param status: Status to set on object
        :return:
        """
        if status not in DirectoryStatus.VALID_STATUS_LIST:
            raise Exception("Invalid status written to DirectoryStatus object.")
        self._status = status

    def status_equals(self, status):
        """
        Checks if status on object is the same as status given.
        Only accepts status's from the predefined VALID_STATUS_LIST
        :param status: check given status against current status
        :return:
        """
        if status not in DirectoryStatus.VALID_STATUS_LIST:
            raise Exception("Invalid status '{}' used for comparison.".format(status))
        return self._status == status

    @property
    def message(self):
        return self._message

    @message.setter
    def message(self, message):
        self._message = message



from os import path, walk
from csv import reader
from collections import OrderedDict
from copy import deepcopy
import logging

import irida_uploader_cl.model as model
from irida_uploader_cl.parsers import exceptions


def build_sequencing_run_from_samples(sample_sheet_file):
    """
    Create a SequencingRun object with full project/sample/sequence_file structure

    :param sample_sheet_file:
    :return: SequencingRun
    """
    sample_list = _parse_sample_list(sample_sheet_file)

    logging.debug("Building SequencingRun from parsed data")

    # create list of projects and add samples to appropriate project
    project_list = []
    for sample_number, sample in enumerate(sample_list):
        # get data from data dict
        sample_name = sample['Sample_Name']
        project_id = sample['Project_ID']
        file_f = sample['File_Forward']
        file_r = sample['File_Reverse']

        project = None
        # see if project exists
        for p in project_list:
            if project_id == p.id:
                project = p
        # create project if it doesn't exitt yet
        if project is None:
            project = model.Project(id=project_id)
            project_list.append(project)

        # create sequence file
        if len(file_r) > 0:
            # paired end read
            sq = model.SequenceFile(properties_dict=None, file_list=[file_f, file_r])
        else:
            # single end read
            sq = model.SequenceFile(properties_dict=None, file_list=[file_f])

        # create sample
        sample_obj = model.Sample(sample_name=sample_name, sample_number=sample_number + 1)

        # add sequence file to sample
        sample_obj.sequence_file = deepcopy(sq)

        # add sample to project
        project.add_sample(sample_obj)

    # add the layout type to the sequencing run so we know if it is paired or single end
    if project_list[0].sample_list[0].sequence_file.is_paired_end():
        metadata = {'layoutType': 'PAIRED_END'}
    else:
        metadata = {'layoutType': 'SINGLE_END'}

    sequence_run = model.SequencingRun(metadata=metadata, project_list=project_list)
    logging.debug("SequencingRun built")
    return sequence_run


def _parse_sample_list(sample_sheet_file):
    """
    Creates a list of all sample data in the sample_sheet_file
    Verifies data is valid for uploading

    :param sample_sheet_file:
    :return: list of sample data dicts
    """
    sample_dict_list = _parse_samples(sample_sheet_file)

    data_dir = path.dirname(sample_sheet_file)
    data_dir_file_list = next(walk(data_dir))[2]  # Create a file list of the data directory, only hit the os once

    data_dir_file_list_full_path = []
    for file_name in data_dir_file_list:
        data_dir_file_list_full_path.append(path.join(path.abspath(data_dir), file_name))
    has_paired_end_read = False
    has_single_end_read = False

    logging.info("Verifying data parsed from sample sheet {}".format(sample_sheet_file))

    for sample_dict in sample_dict_list:

        paired_end_read = len(sample_dict['File_Reverse']) > 0
        # keep track if we have both paired and single end reads
        if paired_end_read:
            has_paired_end_read = True
        else:
            has_single_end_read = True

        # Check if file names are in the files we found in the directory
        if ((sample_dict['File_Forward'] not in data_dir_file_list) and (
                sample_dict['File_Forward'] not in data_dir_file_list_full_path)):
            raise exceptions.SampleSheetError(
                ("Your sample sheet is malformed. {} Does not match any file in the directory {}"
                 "".format(sample_dict['File_Forward'], data_dir)),
                sample_sheet_file
            )
        if ((paired_end_read and sample_dict['File_Reverse'] not in data_dir_file_list) and (
                paired_end_read and sample_dict['File_Reverse'] not in data_dir_file_list_full_path)):
            raise exceptions.SampleSheetError(
                ("Your sample sheet is malformed. {} Does not match any file in the directory {}"
                 "".format(sample_dict['File_Reverse'], data_dir)),
                sample_sheet_file
            )

        # Add the dir to each file to create the full path
        if sample_dict['File_Forward'] not in data_dir_file_list_full_path:
            sample_dict['File_Forward'] = path.join(data_dir, sample_dict['File_Forward'])
        if paired_end_read and sample_dict['File_Reverse'] not in data_dir_file_list_full_path:
            sample_dict['File_Reverse'] = path.join(data_dir, sample_dict['File_Reverse'])

    # Verify we don't have both single end and paired end reads
    if has_single_end_read and has_paired_end_read:
        raise exceptions.SampleSheetError(
            ("Your sample sheet is malformed. "
             "SampleSheet cannot have both paired end and single end runs. "
             "Make sure all samples are either paired or single."),
            sample_sheet_file
        )

    return sample_dict_list


def _parse_samples(sample_sheet_file):

    """
    Parse all the lines under "[Data]" in .csv file

    arguments:
            sample_sheet_file -- path to SampleSheet.csv

    returns a list containing dictionaries with the properties from the csv file
    """

    logging.info("Reading data from sample sheet {}".format(sample_sheet_file))

    csv_reader = get_csv_reader(sample_sheet_file)
    # start with an ordered dictionary so that keys are ordered in the same
    # way that they are inserted.
    sample_dict = OrderedDict()
    sample_dict_list = []

    sample_key_list = ['Sample_Name', 'Project_ID', 'File_Forward', 'File_Reverse']

    # initilize dictionary keys from first line (data headers/attributes)
    set_attributes = False
    for line in csv_reader:

        if set_attributes:
            for item in line:

                if item in sample_key_list:
                    key_name = item
                    sample_dict[key_name] = ""

            break

        if "[Data]" in line:
            set_attributes = True

    # fill in values for keys. line is currently below the [Data] headers
    for sample_number, line in enumerate(csv_reader):
        # if the line is empty (like a blank line at the end of the file) continue
        if not line:
            continue

        if len(sample_dict.keys()) != len(line):
            """
            if there is one more Data header compared to the length of
            data values then add an empty string to the end of data values
            i.e the File_Reverse will be empty string
            assumes the last Data header is going to be the File_Reverse
            this handles the case where the last trailing comma is trimmed when
            doing a single end run
            """
            if len(sample_dict.keys()) - len(line) == 1:
                line.append("")
            else:
                raise exceptions.SampleSheetError(
                    ("Your sample sheet is malformed. Expected to find {} "
                     "columns in the [Data] section, but only found {} columns "
                     "for line {}.".format(len(sample_dict.keys()), len(line), line)),
                    sample_sheet_file
                )

        for index, key in enumerate(sample_dict.keys()):
            value = line[index].strip()

            # Keys other than 'File_Reverse' cannot be empty
            if len(value) is 0:  # no value
                if key != 'File_Reverse':
                    raise exceptions.SampleSheetError(
                        ("Your sample sheet is malformed. {} in the [Data] section cannot be empty."
                         "".format(key)),
                        sample_sheet_file
                    )

            sample_dict[key] = value

        sample_dict_list.append(deepcopy(sample_dict))

    return sample_dict_list


def get_csv_reader(sample_sheet_file):

    """
    tries to create a csv.reader object which will be used to
        parse through the lines in SampleSheet.csv
    raises an error if:
            sample_sheet_file is not an existing file
            sample_sheet_file contains null byte(s)

    arguments:
            data_dir -- the directory that has SampleSheet.csv in it

    returns a csv.reader object
    """

    if path.isfile(sample_sheet_file):
        csv_file = open(sample_sheet_file, "r")
        # strip any trailing newline characters from the end of the line
        # including Windows newline characters (\r\n)
        csv_lines = [x.rstrip('\n') for x in csv_file]
        csv_lines = [x.rstrip('\r') for x in csv_lines]

        # open and read file in binary then send it to be parsed by csv's reader
        csv_reader = reader(csv_lines)
    else:
        raise exceptions.SampleSheetError(
            "Sample sheet cannot be parsed as a CSV file because it's not a regular file.",
            sample_sheet_file)

    return csv_reader



from irida_uploader_cl.progress.upload_status import get_directory_status, write_directory_status
from irida_uploader_cl.progress.upload_signals import signal_worker, send_progress, ProgressData
from irida_uploader_cl.progress import exceptions



import logging

from irida_uploader_cl.parsers import directory, miseq, miniseq, nextseq

supported_parsers = [
    'miseq',
    'miniseq',
    'nextseq',
    'directory',
]


class Parser:
    """
    This class handles creation of the different parser objects.

    When creating a new parser, the parser type can be added here to enable it's usage.

    Like the miseq and directory parser, a new parser class needs the following static methods
        find_single_run(directory)
        find_runs(directory)
        get_required_file_list()
        get_sample_sheet(directory)
        get_sequencing_run(sample_sheet)

    """

    @staticmethod
    def factory(parser_type):
        """
        This factory creates and returns parser objects

        example:
            from parser import Parser
            my_parser = Parser.factory("directory")

        :param parser_type: a String of a valid parser name
        :return:
        """
        if parser_type == "directory":
            logging.debug("Creating directory parser")
            return directory.Parser()
        if parser_type == "miseq":
            logging.debug("Creating miseq parser")
            return miseq.Parser()
        if parser_type == "miniseq":
            logging.debug("Creating miniseq parser")
            return miniseq.Parser()
        if parser_type == "nextseq":
            logging.debug("Creating nextseq parser")
            return nextseq.Parser()
        raise AssertionError("Bad parser creation, invalid parser_type given: {}".format(parser_type))



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name='irida_uploader_cl',
    version='0.2',
    description='IRIDA uploader: upload NGS data to IRIDA system',
    url='https://github.com/duanjunhyq/irida_uploader_cl',
    author='Jun Duan',
    author_email='jun.duan@bccdc.ca',
    long_description=long_description,
    long_description_content_type="text/markdown",
    license='MIT',
    keywords="IRIDA NGS uploader",    
    packages=setuptools.find_packages(include=[ 'irida_uploader_cl',
                                                'irida_uploader_cl.*',
                                                ]),
    install_requires=['rauth',
        'autopep8',
        'pycodestyle',
        'requests',
        'appdirs',
        'pytest',
        'mkdocs',
        'cerberus',
        'argparse',
        'selenium',
        ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    include_package_data = True,
    python_requires='>=3.5',
    entry_points={
    	'console_scripts': [
    		'irida_uploader_cl=irida_uploader_cl.irida_uploader_cl:main'
    	]
    },
  
)



import logging

from pprint import pformat

import irida_uploader_cl.api as api
import irida_uploader_cl.parsers as parsers
import irida_uploader_cl.progress as progress
from irida_uploader_cl.model import DirectoryStatus

from irida_uploader_cl.core import api_handler, parsing_handler, logger, exit_return

VERSION_NUMBER = "0.3.1"


def upload_run_single_entry(directory, force_upload=False):
    """
    This function acts as a single point of entry for uploading a directory

    Handles getting a directories run status, and running if conditions are met (valid run, new run or forced upload).

    :param directory: Directory of the sequencing run to upload
    :param force_upload: When set to true, the upload status file will be ignored and file will attempt to be uploaded
    :return: ExitReturn
    """
    directory_status = parsing_handler.get_run_status(directory)
    # Check if a run is invalid, an invalid run cannot be uploaded.
    if directory_status.status_equals(DirectoryStatus.INVALID):
        error_msg = "ERROR! Run in directory {} is invalid. Returned with message: '{}'".format(
            directory_status.directory, directory_status.message)
        logging.error(error_msg)
        return exit_error(error_msg)

    # Only upload if run is new, or force_upload is True
    if not force_upload:
        if not directory_status.status_equals(DirectoryStatus.NEW):
            error_msg = "ERROR! Run in directory {} is not new. It has either been uploaded, " \
                        "or an upload was attempted with error. " \
                        "Please check the status file 'irida_uploader_status.info' " \
                        "in the run directory for more details. " \
                        "You can bypass this error by uploading with the --force argument.".format(directory)
            logging.error(error_msg)
            return exit_error(error_msg)

    return _validate_and_upload(directory_status)


def batch_upload_single_entry(batch_directory, force_upload=False):
    """
    This function acts as a single point of entry for batch uploading run directories

    It uses _validate_and_upload as it function for uploading the individual runs

    A list of runs to be uploaded is generated at start up, and all found runs will be attempted to be uploaded.

    :param batch_directory: Directory containing sequencing run directories to upload
    :param force_upload: When set to true, the upload status file will be ignored and file will attempt to be uploaded
    :return: ExitReturn
    """
    logging.debug("batch_upload_single_entry:Starting {} with force={}".format(batch_directory, force_upload))
    # get all potential directories to upload
    directory_status_list = parsing_handler.get_run_status_list(batch_directory)
    # list info about directories found
    logging.info("Found {} potential run directories".format(len(directory_status_list)))
    for directory_status in directory_status_list:
        logging.info("DIRECTORY: %s\n"
                     "%30sSTATUS:  %s\n"
                     "%30sDETAILS: %s"
                     % (directory_status.directory, "", directory_status.status, "", directory_status.message))

    # if `force` is on, only don't upload invalid runs
    if force_upload:
        upload_list = [x for x in directory_status_list if not x.status_equals(DirectoryStatus.INVALID)]
        logging.info("Starting upload for all non invalid runs. {} run(s) found. "
                     "(Running with --force)".format(len(upload_list)))
    # without `force` only upload new runs
    else:
        upload_list = [x for x in directory_status_list if x.status_equals(DirectoryStatus.NEW)]
        logging.info("Starting upload for all new runs. {} run(s) found.".format(len(upload_list)))

    # run upload, keep track of which directories did not upload
    error_list = []
    for directory_status in upload_list:
        logging.info("Starting upload for {}".format(directory_status.directory))
        result = _validate_and_upload(directory_status)
        if result.exit_code == exit_return.EXIT_CODE_ERROR:
            error_list.append(directory_status.directory)

    logging.info("Uploads completed with {} error(s)".format(len(error_list)))
    for directory in error_list:
        logging.warning("Directory '{}' upload exited with ERROR, check log and status file for details"
                        "".format(directory))

    logging.info("Batch upload complete, Exiting!")
    return exit_success()


def _validate_and_upload(directory_status):
    """
    This function attempts to upload a single run directory

    Handles parsing and validating the directory for samples
    Sets up the api layer based on config file
    Verifies samples is able to be uploaded (verifies projects exist)
    Initializes objects/routes on IRIDA to accept Samples (creates samples if they don't exist)
    Starts the upload

    :param directory_status: DirectoryStatus object that has directory to try upload
    :return: ExitReturn
    """
    logging_start_block(directory_status.directory)
    logging.debug("upload_run_single_entry:Starting {}".format(directory_status.directory))

    # Add progress file to directory
    try:
        _set_and_write_directory_status(directory_status, DirectoryStatus.PARTIAL)
    except progress.exceptions.DirectoryError as e:
        logging.error("ERROR! Error while trying to write status file to directory {} with error message: {}"
                      "".format(e.directory, e.message))
        logging.info("Samples not uploaded!")
        return exit_error()

    # Do parsing (Also offline validation)
    try:
        sequencing_run = parsing_handler.parse_and_validate(directory_status.directory)
    except parsers.exceptions.DirectoryError as e:
        # Directory was not valid for some reason
        full_error = "ERROR! An error occurred with directory '{}', with message: {}".format(e.directory, e.message)
        logging.error(full_error)
        logging.info("Samples not uploaded!")
        _set_and_write_directory_status(directory_status, DirectoryStatus.ERROR, full_error)
        return exit_error(e)
    except parsers.exceptions.ValidationError as e:
        # Sequencing Run / SampleSheet was not valid for some reason
        error_msg = "ERROR! Errors occurred during validation with message: {}".format(e.message)
        logging.error(error_msg)
        error_list_msg = "Error list: " + pformat(e.validation_result.error_list)
        logging.error(error_list_msg)
        logging.info("Samples not uploaded!")
        full_error = error_msg + ", " + error_list_msg
        _set_and_write_directory_status(directory_status, DirectoryStatus.ERROR, full_error)
        return exit_error(e)

    # Initialize the api for first use
    logging.info("*** Connecting to IRIDA ***")
    try:
        api_handler.initialize_api_from_config()
    except api.exceptions.IridaConnectionError as e:
        logging.error("ERROR! Could not initialize irida api.")
        logging.error("Errors: " + pformat(e.args))
        logging.info("Samples not uploaded!")
        full_error = "ERROR! Could not initialize irida api. Errors: " + pformat(e.args)
        _set_and_write_directory_status(directory_status, DirectoryStatus.ERROR, full_error)
        return exit_error(e)
    logging.info("*** Connected ***")

    logging.info("*** Verifying run (online validation) ***")
    try:
        validation_result = api_handler.prepare_and_validate_for_upload(sequencing_run)
    except api.exceptions.IridaConnectionError as e:
        logging.error("Lost connection to Irida")
        logging.error("Errors: " + pformat(e.args))
        full_error = "Lost connection to Irida. Errors: " + pformat(e.args)
        _set_and_write_directory_status(directory_status, DirectoryStatus.ERROR, full_error)
        return exit_error(e)

    if not validation_result.is_valid():
        logging.error("Sequencing run can not be uploaded")
        logging.error("Sequencing run can not be uploaded. Encountered {} errors"
                      "".format(validation_result.error_count()))
        logging.error("Errors: " + pformat(validation_result.error_list))
        full_error = "Sequencing run can not be uploaded, Errors: " + pformat(validation_result.error_list)
        _set_and_write_directory_status(directory_status, DirectoryStatus.ERROR, full_error)
        return exit_error(full_error)
    logging.info("*** Run Verified ***")

    # Start upload
    logging.info("*** Starting Upload ***")
    try:
        run_id = api_handler.upload_sequencing_run(sequencing_run)
    except api.exceptions.IridaConnectionError as e:
        logging.error("Lost connection to Irida")
        logging.error("Errors: " + pformat(e.args))
        full_error = "Lost connection to Irida. Errors: " + pformat(e.args)
        _set_and_write_directory_status(directory_status, DirectoryStatus.ERROR, full_error)
        return exit_error(e)
    except api.exceptions.IridaResourceError as e:
        logging.error("Could not access IRIDA resource")
        logging.error("Errors: " + pformat(e.args))
        full_error = "Could not access IRIDA resource Errors: " + pformat(e.args)
        _set_and_write_directory_status(directory_status, DirectoryStatus.ERROR, full_error)
        return exit_error(e)
    except api.exceptions.FileError as e:
        logging.error("Could not upload file to IRIDA")
        logging.error("Errors: " + pformat(e.args))
        full_error = "Could not upload file to IRIDA. Errors: " + pformat(e.args)
        _set_and_write_directory_status(directory_status, DirectoryStatus.ERROR, full_error)
        return exit_error(e)
    logging.info("*** Upload Complete ***")

    # Set progress file to complete
    try:
        _set_and_write_directory_status(directory_status, DirectoryStatus.COMPLETE, run_id=run_id)
    except progress.exceptions.DirectoryError as e:
        # this is an exceptionally rare case (successful upload, but fails to write progress)
        logging.ERROR("ERROR! Error while trying to write status file to directory {} with error message: {}"
                      "".format(e.directory, e.message))
        logging.info("Samples were uploaded, but progress file may be incorrect!")

    logging.info("Samples in directory '{}' have finished uploading!".format(directory_status.directory))

    logging_end_block()

    return exit_success()


def _set_and_write_directory_status(directory_status, status, message=None, run_id=None):
    """
    Given a DirectoryStatus object, sets the status and message, and then writes to the directory status directory

    :param directory_status: DirectoryStatus object
    :param status: a valid DirectoryStatus status
    :param message: string
    :param run_id: optional, if provided, the run id and irida instance will be included when written
    :return:
    """
    directory_status.status = status
    directory_status.message = message
    if run_id:
        progress.write_directory_status(directory_status, run_id)
    else:
        progress.write_directory_status(directory_status)


def exit_error(error):
    """
    Returns an failed run exit code which ends the process when returned
    :return: ExitReturn with EXIT_CODE_ERROR
    """
    logging_end_block()
    return exit_return.ExitReturn(exit_return.EXIT_CODE_ERROR, error)


def exit_success():
    """
    Returns an success run exit code which ends the process when returned
    :return: ExitReturn with EXIT_CODE_SUCCESS
    """
    return exit_return.ExitReturn(exit_return.EXIT_CODE_SUCCESS)


def logging_start_block(directory):
    """
    Logs an information block to the console and file which indicates the start of an upload run.
    Includes the uploader version number set in this module
    :return:
    """
    logger.add_log_to_directory(directory)
    logging.info("==================================================")
    logging.info("---------------STARTING UPLOAD RUN----------------")
    logging.info("Uploader Version {}".format(VERSION_NUMBER))
    logging.info("Logging to file in: " + logger.get_user_log_dir())
    logging.info("==================================================")


def logging_end_block():
    """
    Logs an block to the console and file that indicates the end of an upload run.
    :return:
    """
    logging.info("==================================================")
    logging.info("----------------ENDING UPLOAD RUN-----------------")
    logging.info("==================================================")
    logger.remove_directory_logger()



import logging
import os

import irida_uploader_cl.model as model
import irida_uploader_cl.progress as progress

from irida_uploader_cl.parsers import exceptions
from irida_uploader_cl.parsers.nextseq import sample_parser, validation


class Parser:

    SAMPLE_SHEET_FILE_NAME = 'SampleSheet.csv'
    UPLOAD_COMPLETE_FILE_NAME = 'RTAComplete.txt'

    @staticmethod
    def get_required_file_list():
        """
        Returns a list of files that are required for a run directory to be considered valid
        :return: [files_names]
        """
        logging.warning("NOTE: If bcl2fastq has not finished, run may return as invalid, "
                        "or incomplete files could be uploaded!")
        return [
            Parser.SAMPLE_SHEET_FILE_NAME,
            Parser.UPLOAD_COMPLETE_FILE_NAME
        ]

    @staticmethod
    def _find_directory_list(directory):
        """Find and return all directories in the specified directory.

        Arguments:
        directory -- the directory to find directories in

        Returns: a list of directories including current directory
        """

        # Checks if we can access to the given directory, return empty and log a warning if we cannot.
        if not os.access(directory, os.W_OK):
            raise exceptions.DirectoryError("The directory is not writeable, "
                                            "can not upload samples from this directory {}".format(directory),
                                            directory)

        dir_list = next(os.walk(directory))[1]  # Gets the list of directories in the directory
        full_dir_list = []
        for d in dir_list:
            full_dir_list.append(os.path.join(directory, d))
        return full_dir_list

    @staticmethod
    def find_runs(directory):
        """
        find a list of run directories in the directory given

        :param directory:
        :return: list of DirectoryStatus objects
        """
        logging.info("looking for runs in {}".format(directory))

        runs = []
        directory_list = Parser._find_directory_list(directory)
        for d in directory_list:
            runs.append(progress.get_directory_status(d, Parser.get_required_file_list()))

        return runs

    @staticmethod
    def find_single_run(directory):
        """
        Find a run in the base directory given

        :param directory:
        :return: DirectoryStatus object
        """
        logging.info("looking for run in {}".format(directory))

        return progress.get_directory_status(directory, Parser.get_required_file_list())

    @staticmethod
    def get_sample_sheet(directory):
        """
        gets the sample sheet file path from a given run directory

        :param directory:
        :return:
        """
        logging.info("Looking for sample sheet in {}".format(directory))

        # Checks if we can access to the given directory, return empty and log a warning if we cannot.
        if not os.access(directory, os.W_OK):
            logging.error(("The directory is not accessible, can not parse samples from this directory {}"
                           "".format(directory), directory))
            raise exceptions.DirectoryError("The directory is not accessible, "
                                            "can not parse samples from this directory {}".format(directory), directory)

        sample_sheet_file_name = Parser.SAMPLE_SHEET_FILE_NAME
        file_list = next(os.walk(directory))[2]  # Gets the list of files in the directory
        if sample_sheet_file_name not in file_list:
            logging.error("No sample sheet file in the NextSeq format found")
            raise exceptions.DirectoryError("The directory {} has no sample sheet file in the NextSeq format"
                                            " with the name {}"
                                            .format(directory, sample_sheet_file_name), directory)
        else:
            logging.debug("Sample sheet found")
            return os.path.join(directory, sample_sheet_file_name)

    @staticmethod
    def get_sequencing_run(sample_sheet):
        """
        Does local validation on the integrety of the run directory / sample sheet

        Throws a ValidationError with a valadation result attached if it cannot make a sequencing run

        :param sample_sheet:
        :return: SequencingRun
        """

        # Try to get the sample sheet, validate that the sample sheet is valid
        validation_result = validation.validate_sample_sheet(sample_sheet)
        if not validation_result.is_valid():
            logging.error("Errors occurred while getting sample sheet")
            raise exceptions.ValidationError("Errors occurred while getting sample sheet", validation_result)

        # Try to parse the meta data from the sample sheet, throw validation error if errors occur
        validation_result = model.ValidationResult()
        try:
            run_metadata = sample_parser.parse_metadata(sample_sheet)
        except exceptions.SampleSheetError as error:
            validation_result.add_error(error)
            logging.error("Errors occurred while parsing metadata")
            raise exceptions.ValidationError("Errors occurred while parsing metadata", validation_result)

        # Try to build sequencing run from sample sheet & meta data, raise validation error if errors occur
        try:
            sequencing_run = sample_parser.build_sequencing_run_from_samples(sample_sheet, run_metadata)
        except exceptions.SequenceFileError as error:
            validation_result.add_error(error)
            logging.error("Errors occurred while building sequence run from sample sheet")
            raise exceptions.ValidationError("Errors occurred while building sequence run from sample sheet",
                                             validation_result)

        return sequencing_run



from cerberus import Validator, TypeDefinition

from irida_uploader_cl.model.project import Project


class SequencingRun:

    # Define Project as a type for validation
    _project_type = TypeDefinition('project', (Project,), ())
    Validator.types_mapping['project'] = _project_type

    uploadable_schema = {
        '_project_list': {
            'type': 'list',
            'empty': False,  # must have at least 1 project
            'schema': {'type': 'project'}
        },
        '_metadata': {
            'type': 'dict',
            'schema': {
                'layoutType': {
                    'type': 'string',
                    'required': True,
                    'allowed': ['PAIRED_END', 'SINGLE_END']
                }
            }
        }
    }

    def __init__(self, metadata, project_list):
        self._project_list = project_list
        self._metadata = metadata

    @property
    def metadata(self):
        return self._metadata

    @metadata.setter
    def metadata(self, metadata_dict):
        self._metadata = metadata_dict

    @property
    def project_list(self):
        return self._project_list

    @project_list.setter
    def project_list(self, p_list):
        self._project_list = p_list

    def get_dict(self):
        return self.__dict__



"""
A Sample will store (key: value) pairs using a dictionary.

Keys from a sequencer include: 'sampleName','description'

Keys from Irida will include these AND many others
"""
from cerberus import Validator, TypeDefinition
from copy import deepcopy

from irida_uploader_cl.model.sequence_file import SequenceFile


class Sample:

    # Define SequenceFile as a type for validation
    _sample_type = TypeDefinition('sequence_file', (SequenceFile,), ())
    Validator.types_mapping['sequence_file'] = _sample_type

    uploadable_schema = {
        '_sequence_file': {
            'type': 'sequence_file',
            'nullable': False,
            'required': True,
        },
        '_sample_name': {
            'type': 'string',
            'nullable': False,
            'required': True,
            'minlength': 3  # Minimum sample name length is 3
        },
        '_description': {
            'type': 'string',
            'nullable': True,
            'required': False
        },
        '_sample_number': {
            'anyof_type': ['string', 'integer'],
            'nullable': True,
            'required': False
        }
    }

    def __init__(self, sample_name, description='', sample_number=None, samp_dict=None):
        self._sample_name = sample_name
        self._description = description
        self._sample_number = sample_number
        if samp_dict is None:
            samp_dict = {}
        self._sample_dict = dict(samp_dict)
        self._sequence_file = None

    @property
    def sample_name(self):
        return self._sample_name

    @property
    def description(self):
        return self._description

    @property
    def sample_number(self):
        return self._sample_number

    @property
    def sequence_file(self):
        return self._sequence_file

    @sequence_file.setter
    def sequence_file(self, sq):
        self._sequence_file = sq

    def get_irida_id(self):
        if "identifier" in self._sample_dict:
            return self.get("identifier")
        else:
            return None

    def get_uploadable_dict(self):  # formatting for sending to irida when creating a project
        uploadable_dict = deepcopy(self._sample_dict)
        uploadable_dict['sampleName'] = self.sample_name
        uploadable_dict['description'] = self.description
        return uploadable_dict

    def __getitem__(self, key):
        if key in self._sample_dict:
            return self._sample_dict[key]
        return None

    def get(self, key):
        return self.__getitem__(key)

    def __str__(self):
        return str(self.get_uploadable_dict) + str(self.sequence_file)

    def get_dict(self):
        return self.__dict__
